// ConnectionManager.cpp

#include "ConnectionManager.h"
#include <stdio.h>
#include <assert.h>

#define MODE_SERVER	1
#define MODE_CLIENT	2

CConnectionManager::CConnectionManager ()
{
	MsgSize = 0;
	WorkMode = 0;
	ClientArray = NULL;

	strcpy (ErrorMsg, "No error");
}

CConnectionManager::~CConnectionManager ()
{
	if (WorkMode)
	{
		closesocket(Socket);
		WSACleanup();
	}
	if (ClientArray != NULL)
		delete ClientArray;
}

int CConnectionManager::InitServer (unsigned short LocalPort, int ClientsNumber)
{
	assert (WorkMode == 0); // System should not be initialized before

	if (WSAStartup (0x202,&wsaData) == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg,"Windows Socket startup error. Error = %d",WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	Socket = socket (AF_INET, SOCK_DGRAM, 0);
	if (Socket == INVALID_SOCKET)
	{
		wsprintf (ErrorMsg,"Cannot create socket. Error = %d",WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	memset (&LocalAddress, 0, sizeof (LocalAddress));
	LocalAddress.sin_family = AF_INET;
	LocalAddress.sin_port = htons(LocalPort);
	LocalAddress.sin_addr.s_addr = INADDR_ANY;

	if (bind (Socket, (struct sockaddr*)&LocalAddress, sizeof (LocalAddress)) == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg,"bind () failed. Error = %d", WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	this -> LocalPort = LocalPort;
	this -> ClientsNumber = ClientsNumber;
	// Allocate memory for ClientsNumber clients
	if (ClientsNumber > MAX_CLIENT_NUMBER)
	{
		strcpy (ErrorMsg,"Too many clients");
		return -1;
	}

	ClientArray = new sockaddr_in [ClientsNumber];
	memset (ClientArray, 0, sizeof (sockaddr_in) * ClientsNumber);
	ClientCount = 0;

	WorkMode = MODE_SERVER;
	return 0;
}

int CConnectionManager::SendToClient (const char *Buffer, int ClientIndex)
{
	int ReturnValue;
	int Count;
	char XOR = 0;
	char *TempBuffer;

	assert (WorkMode == MODE_SERVER);
	assert (MsgSize > 0);

	if (ClientCount < 1)
	{
		strcpy (ErrorMsg, "No client connected");
		return -1;
	}
	if (ClientIndex >= 0 && ClientIndex < ClientCount)
	{
		TempBuffer = (char *)Buffer;
		for (Count = 0; Count < MsgSize - 1; Count ++)
		{
			XOR ^= TempBuffer [Count];
		}
		TempBuffer [MsgSize - 1] = XOR;
		ReturnValue = sendto (Socket, TempBuffer, MsgSize, 0, (struct sockaddr *)&ClientArray [ClientIndex], sizeof (sockaddr_in));
		if (ReturnValue == SOCKET_ERROR)
		{
			wsprintf (ErrorMsg, "SendTo failed. Error = %d", WSAGetLastError ());
			return -1;
		}
	}
	else
	{
		strcpy (ErrorMsg, "Invalid client");
		return -1;
	}
	return 0;
}

int CConnectionManager::SendAllClients (const char *Buffer, int Except)
{
	int Count;

	assert (WorkMode == MODE_SERVER);
	assert (MsgSize > 0);

	if (Except != -1)
	{
		for (Count = 0; Count < ClientCount; Count ++)
		{
			if (Except == Count)
				continue;
			SendToClient (Buffer, Count);
		}
	}
	else
	{
		for (Count = 0; Count < ClientCount; Count ++)
		{
			SendToClient (Buffer, Count);
		}
	}
	return 0;
}

int CConnectionManager::SendLastClient (const char *Buffer)
{
	int ReturnValue;
	int Count;
	char XOR = 0;
	char *TempBuffer;

	assert (WorkMode == MODE_SERVER);
	assert (MsgSize > 0);
 
	TempBuffer = (char *)Buffer;
	for (Count = 0; Count < MsgSize - 1; Count ++)
	{
		XOR ^= TempBuffer [Count];
	}
	TempBuffer [MsgSize - 1] = XOR;
	ReturnValue = sendto (Socket, TempBuffer, MsgSize, 0, (struct sockaddr *)&LastClient, sizeof (sockaddr_in));
	if (ReturnValue == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg, "SendTo failed. Error = %d", WSAGetLastError ());
		return -1;
	}
	return 0;
}

int CConnectionManager::ReceiveFromClient (char *Buffer, int *ClientIndex)
{
	int ReturnValue;
	fd_set Sets;
	timeval TimeVal = {0, 0};
	sockaddr_in From;
	int FromSize;
	int Count;
	char XOR = 0;

	assert (WorkMode == MODE_SERVER);
	assert (MsgSize > 0);

	FD_ZERO (&Sets);
	FD_SET (Socket, &Sets);
	ReturnValue = select (0, &Sets, NULL, NULL, &TimeVal);
	if (ReturnValue == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg, "Select failed. Error = %d", WSAGetLastError ());
		return -1;
	}
	if (FD_ISSET (Socket, &Sets) != 0 && ReturnValue > 0) // There is data to read
	{
		FromSize = sizeof (From);
		recvfrom (Socket, Buffer, MsgSize, 0, (struct sockaddr *)&From, &FromSize);

		// Check Buffer's integrity
		for (Count = 0; Count < MsgSize - 1; Count ++)
		{
			XOR ^= Buffer [Count];
		}
		if (XOR != Buffer [MsgSize - 1])
		{
			strcpy (ErrorMsg, "Message corrupted");
			Buffer = NULL;
			*ClientIndex = -1;
			return -1;
		}

		LastClient = From;

		// Check if this client is already in ClientArray
		for (Count = 0; Count < ClientCount; Count ++)
		{
			if ((ClientArray [Count].sin_addr.S_un.S_addr == From.sin_addr.S_un.S_addr) && 
				(ClientArray [Count].sin_port == From.sin_port))
			{
				*ClientIndex = Count;
				return 0;
			}
		}
		*ClientIndex = ClientCount;
		return 0;
	}
	else
	{
		return CONNECTION_NO_DATA;
	}
}

int CConnectionManager::AcceptLastClient ()
{
	assert (WorkMode == MODE_SERVER);

	if (ClientCount < ClientsNumber)
	{
		ClientArray [ClientCount] = LastClient;
		ClientCount ++;
		return 0;
	}
	else
	{
		strcpy (ErrorMsg, "Cannot add more clients");
		return -1;
	}
}

int CConnectionManager::DeleteClient (int ClientIndex)
{
	int Count;

	assert (WorkMode == MODE_SERVER);

	if (ClientIndex < 0 || ClientIndex >= ClientCount)
	{
		strcpy (ErrorMsg, "Invalid client");
		return -1;
	}

	for (Count = ClientIndex; Count < ClientCount - 1; Count ++)
	{
		ClientArray [Count] = ClientArray [Count + 1];
	}
	ClientCount --;

	return 0;
}

int CConnectionManager::InitClient (const char *IP, unsigned short LocalPort, unsigned short RemotePort)
{
	unsigned Address;

	assert (WorkMode == 0); // System should not be initialized before
	
	if (WSAStartup(0x202,&wsaData) == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg, "WSAStartup failed. Error = %d", WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	Address = inet_addr (IP);
	Host = gethostbyaddr((char *)&Address, 4, AF_INET);
	if (Host == NULL )
	{
		wsprintf(ErrorMsg,"Cannot resolve address [%s]. Error = %d", IP, WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	Socket = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (Socket < 0)
	{
		wsprintf (ErrorMsg, "Error Opening socket. Error = %d", WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	memset (&LocalAddress, 0, sizeof (LocalAddress));
	LocalAddress.sin_family = AF_INET;
	LocalAddress.sin_port = htons (LocalPort);
	LocalAddress.sin_addr.s_addr = INADDR_ANY;

	if (bind (Socket, (struct sockaddr*)&LocalAddress, sizeof (LocalAddress)) == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg,"bind () failed. Error = %d", WSAGetLastError ());
		WSACleanup ();
		return -1;
	}

	LocalAddress.sin_port = htons (RemotePort);
	LocalAddress.sin_addr.s_addr = inet_addr (IP);

	this -> LocalPort = LocalPort;
	this -> RemotePort = RemotePort;
	WorkMode = MODE_CLIENT;
	return 0;
}

int CConnectionManager::SendToServer (const char *Buffer)
{
	char *TempBuffer;
	int Count;
	char XOR = 0;

	assert (WorkMode == MODE_CLIENT);
	assert (MsgSize > 0);

	TempBuffer = (char *)Buffer;
	for (Count = 0; Count < MsgSize - 1; Count ++)
	{
		XOR ^= TempBuffer [Count];
	}
	TempBuffer [MsgSize - 1] = XOR;
	if (sendto (Socket, TempBuffer, MsgSize, 0, (struct sockaddr *)&LocalAddress, sizeof (LocalAddress)) == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg, "Send failed. Error = %d", WSAGetLastError ());
		return -1;
	}
	return 0;
}

int CConnectionManager::ReceiveFromServer (char *Buffer)
{
	int ReturnValue;
	fd_set Sets;
	timeval TimeVal = {0, 0};
	sockaddr_in From;
	int FromSize;
	int Count;
	char XOR = 0;

	assert (WorkMode == MODE_CLIENT);
	assert (MsgSize > 0);

	FD_ZERO (&Sets);
	FD_SET (Socket, &Sets);
	ReturnValue = select (0, &Sets, NULL, NULL, &TimeVal);
	if (ReturnValue == SOCKET_ERROR)
	{
		wsprintf (ErrorMsg, "Select failed. Error = %d", WSAGetLastError ());
		return -1;
	}

	if (FD_ISSET (Socket, &Sets) != 0 && ReturnValue > 0)
	{
		// There is data to read
		FromSize = sizeof (From);
		ReturnValue = recvfrom (Socket, Buffer, MsgSize, 0, (struct sockaddr *)&From, &FromSize);
		if (ReturnValue == SOCKET_ERROR)
		{
			wsprintf (ErrorMsg, "Recv failed. Error = %d", WSAGetLastError ());
			return -1;
		}
		for (Count = 0; Count < MsgSize - 1; Count ++)
		{
			XOR ^= Buffer [Count];
		}
		if (XOR != Buffer [MsgSize - 1])
		{
			strcpy (ErrorMsg, "Message corrupted");
			Buffer = NULL;
			return -1;
		}
		return 0;
	}
	else
	{
		return CONNECTION_NO_DATA;
	}
}

void CConnectionManager::SetMsgSize (int Size)
{
	MsgSize = Size;
}

char *CConnectionManager::GetErrorMsg ()
{
	return ErrorMsg;
}

void CConnectionManager::ShutDown ()
{
	if (WorkMode != 0)
	{
		MsgSize = 0;
		WorkMode = 0;
		if (ClientArray != NULL)
		{
			delete ClientArray;
			ClientArray = NULL;
		}

		strcpy (ErrorMsg, "No error");

		closesocket (Socket);
		WSACleanup ();
	}
}

int CConnectionManager::GetClientIP (char *IP, int ClientId)
{
	assert (WorkMode == MODE_SERVER);

	if (ClientId >= 0 && ClientId < ClientCount)
	{
		wsprintf (IP, "%d.%d.%d.%d", ClientArray [ClientId].sin_addr.S_un.S_un_b.s_b1,
																	ClientArray [ClientId].sin_addr.S_un.S_un_b.s_b2,
																	ClientArray [ClientId].sin_addr.S_un.S_un_b.s_b3,
																	ClientArray [ClientId].sin_addr.S_un.S_un_b.s_b4);
		return 0;
	}
	else
	{
		IP = NULL;
		strcpy (ErrorMsg, "Invalid client");
		return -1;
	}
}